@@ -348,7 +348,7 @@ test_button_event(struct litest_device *dev, unsigned int button, int state)
 {
 	struct libinput *li = dev->libinput;
 
-	litest_button_click(dev, button, state);
+	litest_button_click_debounced(dev, li, button, state);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 
 	litest_assert_button_event(li, button,
@@ -500,7 +500,7 @@ START_TEST(pointer_recover_from_lost_button_count)
 
 	litest_drain_events(dev->libinput);
 
-	litest_button_click(dev, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
 
 	litest_assert_button_event(li,
 				   BTN_LEFT,
@@ -508,15 +508,15 @@ START_TEST(pointer_recover_from_lost_button_count)
 
 	/* Grab for the release to make libinput lose count */
 	libevdev_grab(evdev, LIBEVDEV_GRAB);
-	litest_button_click(dev, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
 	libevdev_grab(evdev, LIBEVDEV_UNGRAB);
 
 	litest_assert_empty_queue(li);
 
-	litest_button_click(dev, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
 	litest_assert_empty_queue(li);
 
-	litest_button_click(dev, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
 	litest_assert_button_event(li,
 				   BTN_LEFT,
 				   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -802,7 +802,10 @@ START_TEST(pointer_seat_button_count)
 	}
 
 	for (i = 0; i < num_devices; ++i)
-		litest_button_click(devices[i], BTN_LEFT, true);
+		litest_button_click_debounced(devices[i],
+					      libinput,
+					      BTN_LEFT,
+					      true);
 
 	libinput_dispatch(libinput);
 	while ((ev = libinput_get_event(libinput))) {
@@ -832,7 +835,10 @@ START_TEST(pointer_seat_button_count)
 	ck_assert_int_eq(seat_button_count, num_devices);
 
 	for (i = 0; i < num_devices; ++i)
-		litest_button_click(devices[i], BTN_LEFT, false);
+		litest_button_click_debounced(devices[i],
+					      libinput,
+					      BTN_LEFT,
+					      false);
 
 	libinput_dispatch(libinput);
 	while ((ev = libinput_get_event(libinput))) {
@@ -921,8 +927,8 @@ START_TEST(pointer_left_handed)
 	ck_assert_int_eq(status, LIBINPUT_CONFIG_STATUS_SUCCESS);
 
 	litest_drain_events(li);
-	litest_button_click(dev, BTN_LEFT, 1);
-	litest_button_click(dev, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
 
 	litest_assert_button_event(li,
 				   BTN_RIGHT,
@@ -931,8 +937,8 @@ START_TEST(pointer_left_handed)
 				   BTN_RIGHT,
 				   LIBINPUT_BUTTON_STATE_RELEASED);
 
-	litest_button_click(dev, BTN_RIGHT, 1);
-	litest_button_click(dev, BTN_RIGHT, 0);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 1);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 0);
 	litest_assert_button_event(li,
 				   BTN_LEFT,
 				   LIBINPUT_BUTTON_STATE_PRESSED);
@@ -941,8 +947,8 @@ START_TEST(pointer_left_handed)
 				   LIBINPUT_BUTTON_STATE_RELEASED);
 
 	if (libinput_device_pointer_has_button(d, BTN_MIDDLE)) {
-		litest_button_click(dev, BTN_MIDDLE, 1);
-		litest_button_click(dev, BTN_MIDDLE, 0);
+		litest_button_click_debounced(dev, li, BTN_MIDDLE, 1);
+		litest_button_click_debounced(dev, li, BTN_MIDDLE, 0);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
@@ -961,14 +967,14 @@ START_TEST(pointer_left_handed_during_click)
 	enum libinput_config_status status;
 
 	litest_drain_events(li);
-	litest_button_click(dev, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
 	libinput_dispatch(li);
 
 	/* Change while button is down, expect correct release event */
 	status = libinput_device_config_left_handed_set(d, 1);
 	ck_assert_int_eq(status, LIBINPUT_CONFIG_STATUS_SUCCESS);
 
-	litest_button_click(dev, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
 
 	litest_assert_button_event(li,
 				   BTN_LEFT,
@@ -992,16 +998,16 @@ START_TEST(pointer_left_handed_during_click_multiple_buttons)
 	litest_disable_middleemu(dev);
 
 	litest_drain_events(li);
-	litest_button_click(dev, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
 	libinput_dispatch(li);
 
 	status = libinput_device_config_left_handed_set(d, 1);
 	ck_assert_int_eq(status, LIBINPUT_CONFIG_STATUS_SUCCESS);
 
 	/* No left-handed until all buttons were down */
-	litest_button_click(dev, BTN_RIGHT, 1);
-	litest_button_click(dev, BTN_RIGHT, 0);
-	litest_button_click(dev, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 1);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
 
 	litest_assert_button_event(li,
 				   BTN_LEFT,
@@ -1102,7 +1108,7 @@ START_TEST(pointer_scroll_button_no_event_before_timeout)
 						 BTN_LEFT);
 	litest_drain_events(li);
 
-	litest_button_click(device, BTN_LEFT, true);
+	litest_button_click_debounced(device, li, BTN_LEFT, true);
 	litest_assert_empty_queue(li);
 
 	for (i = 0; i < 10; i++) {
@@ -1113,7 +1119,7 @@ START_TEST(pointer_scroll_button_no_event_before_timeout)
 
 	litest_timeout_buttonscroll();
 	libinput_dispatch(li);
-	litest_button_click(device, BTN_LEFT, false);
+	litest_button_click_debounced(device, li, BTN_LEFT, false);
 
 	litest_assert_button_event(li, BTN_LEFT,
 				   LIBINPUT_BUTTON_STATE_PRESSED);
@@ -1146,8 +1152,8 @@ START_TEST(pointer_scroll_button_middle_emulation)
 
 	litest_drain_events(li);
 
-	litest_button_click(dev, BTN_LEFT, 1);
-	litest_button_click(dev, BTN_RIGHT, 1);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 1);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 1);
 	libinput_dispatch(li);
 	litest_timeout_buttonscroll();
 	libinput_dispatch(li);
@@ -1159,8 +1165,8 @@ START_TEST(pointer_scroll_button_middle_emulation)
 
 	libinput_dispatch(li);
 
-	litest_button_click(dev, BTN_LEFT, 0);
-	litest_button_click(dev, BTN_RIGHT, 0);
+	litest_button_click_debounced(dev, li, BTN_LEFT, 0);
+	litest_button_click_debounced(dev, li, BTN_RIGHT, 0);
 	libinput_dispatch(li);
 
 	litest_assert_scroll(li, LIBINPUT_POINTER_AXIS_SCROLL_VERTICAL, -1);
@@ -1525,16 +1531,16 @@ START_TEST(middlebutton)
 	litest_drain_events(li);
 
 	for (i = 0; i < ARRAY_LENGTH(btn); i++) {
-		litest_button_click(device, btn[i][0], true);
-		litest_button_click(device, btn[i][1], true);
+		litest_button_click_debounced(device, li, btn[i][0], true);
+		litest_button_click_debounced(device, li, btn[i][1], true);
 
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 		litest_assert_empty_queue(li);
 
-		litest_button_click(device, btn[i][2], false);
-		litest_button_click(device, btn[i][3], false);
+		litest_button_click_debounced(device, li, btn[i][2], false);
+		litest_button_click_debounced(device, li, btn[i][3], false);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -1568,33 +1574,33 @@ START_TEST(middlebutton_nostart_while_down)
 	if (status == LIBINPUT_CONFIG_STATUS_UNSUPPORTED)
 		return;
 
-	litest_button_click(device, BTN_MIDDLE, true);
+	litest_button_click_debounced(device, li, BTN_MIDDLE, true);
 	litest_drain_events(li);
 
 	for (i = 0; i < ARRAY_LENGTH(btn); i++) {
-		litest_button_click(device, btn[i][0], true);
+		litest_button_click_debounced(device, li, btn[i][0], true);
 		litest_assert_button_event(li,
 					   btn[i][0],
 					   LIBINPUT_BUTTON_STATE_PRESSED);
-		litest_button_click(device, btn[i][1], true);
+		litest_button_click_debounced(device, li, btn[i][1], true);
 		litest_assert_button_event(li,
 					   btn[i][1],
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 
 		litest_assert_empty_queue(li);
 
-		litest_button_click(device, btn[i][2], false);
+		litest_button_click_debounced(device, li, btn[i][2], false);
 		litest_assert_button_event(li,
 					   btn[i][2],
 					   LIBINPUT_BUTTON_STATE_RELEASED);
-		litest_button_click(device, btn[i][3], false);
+		litest_button_click_debounced(device, li, btn[i][3], false);
 		litest_assert_button_event(li,
 					   btn[i][3],
 					   LIBINPUT_BUTTON_STATE_RELEASED);
 		litest_assert_empty_queue(li);
 	}
 
-	litest_button_click(device, BTN_MIDDLE, false);
+	litest_button_click_debounced(device, li, BTN_MIDDLE, false);
 	litest_drain_events(li);
 }
 END_TEST
@@ -1616,7 +1622,7 @@ START_TEST(middlebutton_timeout)
 
 	for (button = BTN_LEFT; button <= BTN_RIGHT; button++) {
 		litest_drain_events(li);
-		litest_button_click(device, button, true);
+		litest_button_click_debounced(device, li, button, true);
 		litest_assert_empty_queue(li);
 		litest_timeout_middlebutton();
 
@@ -1624,7 +1630,7 @@ START_TEST(middlebutton_timeout)
 					   button,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 
-		litest_button_click(device, button, false);
+		litest_button_click_debounced(device, li, button, false);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -1657,22 +1663,22 @@ START_TEST(middlebutton_doubleclick)
 	litest_drain_events(li);
 
 	for (i = 0; i < ARRAY_LENGTH(btn); i++) {
-		litest_button_click(device, btn[i][0], true);
-		litest_button_click(device, btn[i][1], true);
+		litest_button_click_debounced(device, li, btn[i][0], true);
+		litest_button_click_debounced(device, li, btn[i][1], true);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 		litest_assert_empty_queue(li);
 
-		litest_button_click(device, btn[i][2], false);
-		litest_button_click(device, btn[i][2], true);
+		litest_button_click_debounced(device, li, btn[i][2], false);
+		litest_button_click_debounced(device, li, btn[i][2], true);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
-		litest_button_click(device, btn[i][3], false);
+		litest_button_click_debounced(device, li, btn[i][3], false);
 
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
@@ -1705,8 +1711,8 @@ START_TEST(middlebutton_middleclick)
 	for (button = BTN_LEFT; button <= BTN_RIGHT; button++) {
 		/* release button before middle */
 		litest_drain_events(li);
-		litest_button_click(device, button, true);
-		litest_button_click(device, BTN_MIDDLE, true);
+		litest_button_click_debounced(device, li, button, true);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, true);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
@@ -1714,19 +1720,19 @@ START_TEST(middlebutton_middleclick)
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 		litest_assert_empty_queue(li);
-		litest_button_click(device, button, false);
+		litest_button_click_debounced(device, li, button, false);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
-		litest_button_click(device, BTN_MIDDLE, false);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, false);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
 		litest_assert_empty_queue(li);
 
 		/* release middle before button */
-		litest_button_click(device, button, true);
-		litest_button_click(device, BTN_MIDDLE, true);
+		litest_button_click_debounced(device, li, button, true);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, true);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
@@ -1734,11 +1740,11 @@ START_TEST(middlebutton_middleclick)
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 		litest_assert_empty_queue(li);
-		litest_button_click(device, BTN_MIDDLE, false);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, false);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
-		litest_button_click(device, button, false);
+		litest_button_click_debounced(device, li, button, false);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -1770,14 +1776,14 @@ START_TEST(middlebutton_middleclick_during)
 
 	/* trigger emulation, then real middle */
 	for (button = BTN_LEFT; button <= BTN_RIGHT; button++) {
-		litest_button_click(device, BTN_LEFT, true);
-		litest_button_click(device, BTN_RIGHT, true);
+		litest_button_click_debounced(device, li, BTN_LEFT, true);
+		litest_button_click_debounced(device, li, BTN_RIGHT, true);
 
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 
-		litest_button_click(device, BTN_MIDDLE, true);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, true);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -1788,23 +1794,23 @@ START_TEST(middlebutton_middleclick_during)
 		litest_assert_empty_queue(li);
 
 		/* middle still down, release left/right */
-		litest_button_click(device, button, false);
+		litest_button_click_debounced(device, li, button, false);
 		litest_assert_empty_queue(li);
-		litest_button_click(device, button, true);
+		litest_button_click_debounced(device, li, button, true);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 		litest_assert_empty_queue(li);
 
 		/* release both */
-		litest_button_click(device, BTN_LEFT, false);
-		litest_button_click(device, BTN_RIGHT, false);
+		litest_button_click_debounced(device, li, BTN_LEFT, false);
+		litest_button_click_debounced(device, li, BTN_RIGHT, false);
 		litest_assert_button_event(li,
 					   button,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
 		litest_assert_empty_queue(li);
 
-		litest_button_click(device, BTN_MIDDLE, false);
+		litest_button_click_debounced(device, li, BTN_MIDDLE, false);
 		litest_assert_button_event(li,
 					   BTN_MIDDLE,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -2104,91 +2110,176 @@ START_TEST(pointer_time_usec)
 }
 END_TEST
 
-START_TEST(debounce)
+START_TEST(debounce_bounce)
 {
 	struct litest_device *dev = litest_current_device();
 	struct libinput *li = dev->libinput;
+	unsigned int button = _i; /* ranged test */
+
+	if (!libinput_device_pointer_has_button(dev->libinput_device,
+						button))
+		return;
 
 	litest_disable_middleemu(dev);
+	disable_button_scrolling(dev);
 	litest_drain_events(li);
 
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_KEY, button, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_KEY, button, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-
-	/* expect debouncing on now, this event is ignored */
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_KEY, button, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-
+	libinput_dispatch(li);
+	litest_timeout_debounce();
 	libinput_dispatch(li);
 
 	litest_assert_button_event(li,
-				   BTN_LEFT,
+				   button,
 				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_empty_queue(li);
+
+	litest_event(dev, EV_KEY, button, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, button, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, button, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
 	litest_assert_button_event(li,
-				   BTN_LEFT,
+				   button,
 				   LIBINPUT_BUTTON_STATE_RELEASED);
 
 	litest_assert_empty_queue(li);
 }
 END_TEST
 
-START_TEST(debounce_timer)
+START_TEST(debounce_bounce_check_immediate)
 {
 	struct litest_device *dev = litest_current_device();
 	struct libinput *li = dev->libinput;
 
 	litest_disable_middleemu(dev);
+	disable_button_scrolling(dev);
 	litest_drain_events(li);
 
+	/* Press must be sent without delay */
 	litest_event(dev, EV_KEY, BTN_LEFT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_timeout_debounce();
+	litest_assert_empty_queue(li);
+
+	/* held down & past timeout, we expect releases to be immediate */
 
-	/* expect debouncing on now, this event is ignored */
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 	litest_event(dev, EV_KEY, BTN_LEFT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
 
 	litest_timeout_debounce();
+	litest_assert_empty_queue(li);
+}
+END_TEST
+
+/* Triggers the event sequence that initializes the spurious
+ * debouncing behavior */
+static inline void
+debounce_trigger_spurious(struct litest_device *dev, struct libinput *li)
+{
+	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+
+	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+
+	/* gets filtered now */
+	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
+	litest_assert_empty_queue(li);
+}
+
+START_TEST(debounce_spurious)
+{
+	struct litest_device *dev = litest_current_device();
+	struct libinput *li = dev->libinput;
+	unsigned int button = _i; /* ranged test */
+
+	if (!libinput_device_pointer_has_button(dev->libinput_device,
+						button))
+		return;
+
+	litest_disable_middleemu(dev);
+	disable_button_scrolling(dev);
 	litest_drain_events(li);
 
+	debounce_trigger_spurious(dev, li);
+
 	for (int i = 0; i < 3; i++) {
-		litest_event(dev, EV_KEY, BTN_LEFT, 1);
+		litest_event(dev, EV_KEY, button, 1);
 		litest_event(dev, EV_SYN, SYN_REPORT, 0);
 		libinput_dispatch(li);
 		litest_timeout_debounce();
+		libinput_dispatch(li);
 
 		/* Not all devices can disable middle button emulation, time out on
 		 * middle button here to make sure the initial button press event
 		 * was flushed.
 		 */
-		libinput_dispatch(li);
 		litest_timeout_middlebutton();
 		libinput_dispatch(li);
+
 		litest_assert_button_event(li,
-					   BTN_LEFT,
+					   button,
 					   LIBINPUT_BUTTON_STATE_PRESSED);
 
 		/* bouncy bouncy bouncy */
-		litest_event(dev, EV_KEY, BTN_LEFT, 0);
+		litest_event(dev, EV_KEY, button, 0);
 		litest_event(dev, EV_SYN, SYN_REPORT, 0);
-		litest_event(dev, EV_KEY, BTN_LEFT, 1);
+		litest_event(dev, EV_KEY, button, 1);
 		litest_event(dev, EV_SYN, SYN_REPORT, 0);
 		litest_assert_empty_queue(li);
 
-		litest_event(dev, EV_KEY, BTN_LEFT, 0);
+		litest_event(dev, EV_KEY, button, 0);
 		litest_event(dev, EV_SYN, SYN_REPORT, 0);
 		libinput_dispatch(li);
 		litest_timeout_debounce();
 		libinput_dispatch(li);
 		litest_assert_button_event(li,
-					   BTN_LEFT,
+					   button,
 					   LIBINPUT_BUTTON_STATE_RELEASED);
 
 		litest_assert_empty_queue(li);
@@ -2196,7 +2287,7 @@ START_TEST(debounce_timer)
 }
 END_TEST
 
-START_TEST(debounce_multibounce)
+START_TEST(debounce_spurious_multibounce)
 {
 	struct litest_device *dev = litest_current_device();
 	struct libinput *li = dev->libinput;
@@ -2204,15 +2295,7 @@ START_TEST(debounce_multibounce)
 	litest_disable_middleemu(dev);
 	litest_drain_events(li);
 
-	/* enable debouncing */
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	debounce_trigger_spurious(dev, li);
 	litest_drain_events(li);
 
 	/* Let's assume our button has ventricular fibrilation and sends a
@@ -2223,6 +2306,8 @@ START_TEST(debounce_multibounce)
 
 	litest_event(dev, EV_KEY, BTN_LEFT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
 
 	/* Not all devices can disable middle button emulation, time out on
 	 * middle button here to make sure the initial button press event
@@ -2261,20 +2346,34 @@ START_TEST(debounce_multibounce)
 }
 END_TEST
 
-START_TEST(debounce_no_debounce_for_otherbutton)
+START_TEST(debounce_spurious_dont_enable_on_otherbutton)
 {
 	struct litest_device *dev = litest_current_device();
+	struct libinput_device *device = dev->libinput_device;
 	struct libinput *li = dev->libinput;
 
+	if (!libinput_device_config_middle_emulation_is_available(device))
+		return;
+
 	litest_disable_middleemu(dev);
+	disable_button_scrolling(dev);
 	litest_drain_events(li);
 
+	/* Don't trigger spurious debouncing on otherbutton events */
 	litest_event(dev, EV_KEY, BTN_LEFT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
 	litest_event(dev, EV_KEY, BTN_LEFT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 	litest_event(dev, EV_KEY, BTN_RIGHT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 	litest_event(dev, EV_KEY, BTN_RIGHT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 
@@ -2290,43 +2389,68 @@ START_TEST(debounce_no_debounce_for_otherbutton)
 	litest_assert_button_event(li,
 				   BTN_RIGHT,
 				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
 	litest_assert_button_event(li,
 				   BTN_RIGHT,
 				   LIBINPUT_BUTTON_STATE_RELEASED);
 
 	litest_assert_empty_queue(li);
+
+	/* Expect release to be immediate */
+	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
+	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
 }
 END_TEST
 
-START_TEST(debounce_cancel_debounce_otherbutton)
+START_TEST(debounce_spurious_cancel_debounce_otherbutton)
 {
 	struct litest_device *dev = litest_current_device();
+	struct libinput_device *device = dev->libinput_device;
 	struct libinput *li = dev->libinput;
 
+	if (!libinput_device_config_middle_emulation_is_available(device))
+		return;
+
 	litest_disable_middleemu(dev);
+	disable_button_scrolling(dev);
 	litest_drain_events(li);
 
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-
-	litest_drain_events(li);
+	debounce_trigger_spurious(dev, li);
 
 	litest_event(dev, EV_KEY, BTN_LEFT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
+	/* spurious debouncing is on but the release should get flushed by
+	 * the other button */
 	litest_event(dev, EV_KEY, BTN_LEFT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	/* release is now held back, press was ignored,
-	 * other button should flush the release */
 	litest_event(dev, EV_KEY, BTN_RIGHT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, BTN_LEFT, 1);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	litest_event(dev, EV_KEY, BTN_LEFT, 0);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 	litest_event(dev, EV_KEY, BTN_RIGHT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
 
@@ -2342,6 +2466,12 @@ START_TEST(debounce_cancel_debounce_otherbutton)
 	litest_assert_button_event(li,
 				   BTN_RIGHT,
 				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_PRESSED);
+	litest_assert_button_event(li,
+				   BTN_LEFT,
+				   LIBINPUT_BUTTON_STATE_RELEASED);
 	litest_assert_button_event(li,
 				   BTN_RIGHT,
 				   LIBINPUT_BUTTON_STATE_RELEASED);
@@ -2350,31 +2480,28 @@ START_TEST(debounce_cancel_debounce_otherbutton)
 }
 END_TEST
 
-START_TEST(debounce_switch_to_otherbutton)
+START_TEST(debounce_spurious_switch_to_otherbutton)
 {
 	struct litest_device *dev = litest_current_device();
+	struct libinput_device *device = dev->libinput_device;
 	struct libinput *li = dev->libinput;
 
-	litest_drain_events(li);
-
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 0);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	if (!libinput_device_config_middle_emulation_is_available(device))
+		return;
 
 	litest_drain_events(li);
+	debounce_trigger_spurious(dev, li);
 
 	litest_event(dev, EV_KEY, BTN_LEFT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+	libinput_dispatch(li);
+	litest_timeout_debounce();
+	libinput_dispatch(li);
+
+
 	litest_event(dev, EV_KEY, BTN_LEFT, 0);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	litest_event(dev, EV_KEY, BTN_LEFT, 1);
-	litest_event(dev, EV_SYN, SYN_REPORT, 0);
-	/* release is now held back, press was ignored,
+	/* release is now held back,
 	 * other button should flush the release */
 	litest_event(dev, EV_KEY, BTN_RIGHT, 1);
 	litest_event(dev, EV_SYN, SYN_REPORT, 0);
@@ -2412,6 +2539,7 @@ litest_setup_tests_pointer(void)
 {
 	struct range axis_range = {ABS_X, ABS_Y + 1};
 	struct range compass = {0, 7}; /* cardinal directions */
+	struct range buttons = {BTN_LEFT, BTN_TASK + 1};
 
 	litest_add("pointer:motion", pointer_motion_relative, LITEST_RELATIVE, LITEST_POINTINGSTICK);
 	litest_add_for_device("pointer:motion", pointer_motion_relative_zero, LITEST_MOUSE);
@@ -2474,10 +2602,11 @@ litest_setup_tests_pointer(void)
 
 	litest_add("pointer:time", pointer_time_usec, LITEST_RELATIVE, LITEST_ANY);
 
-	litest_add("pointer:debounce", debounce, LITEST_BUTTON, LITEST_TOUCHPAD);
-	litest_add("pointer:debounce", debounce_timer, LITEST_BUTTON, LITEST_TOUCHPAD);
-	litest_add("pointer:debounce", debounce_multibounce, LITEST_BUTTON, LITEST_TOUCHPAD);
-	litest_add("pointer:debounce_otherbutton", debounce_no_debounce_for_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
-	litest_add("pointer:debounce_otherbutton", debounce_cancel_debounce_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
-	litest_add("pointer:debounce_otherbutton", debounce_switch_to_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
+	litest_add_ranged("pointer:debounce", debounce_bounce, LITEST_BUTTON, LITEST_TOUCHPAD, &buttons);
+	litest_add("pointer:debounce", debounce_bounce_check_immediate, LITEST_BUTTON, LITEST_TOUCHPAD);
+	litest_add_ranged("pointer:debounce", debounce_spurious, LITEST_BUTTON, LITEST_TOUCHPAD, &buttons);
+	litest_add("pointer:debounce", debounce_spurious_multibounce, LITEST_BUTTON, LITEST_TOUCHPAD);
+	litest_add("pointer:debounce_otherbutton", debounce_spurious_dont_enable_on_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
+	litest_add("pointer:debounce_otherbutton", debounce_spurious_cancel_debounce_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
+	litest_add("pointer:debounce_otherbutton", debounce_spurious_switch_to_otherbutton, LITEST_BUTTON, LITEST_TOUCHPAD);
 }
